# -*- coding: utf-8 -*-
"""This module contains the data model of the application."""

# symbols which are imported by "from uzi.model import *"
__all__ = ['Doctor', 'Organ', 'USG_Result', 'USG_SubResult',
    'Group', 'Permission', 'User', 'Visit', 'VisitIdentity']

from datetime import datetime

import pkg_resources
pkg_resources.require("SQLAlchemy >= 0.4.3")
pkg_resources.require("Elixir >= 0.6.1")

# import the basic Elixir classes and functions for declaring the data model
# (see http://elixir.ematia.de/trac/wiki/TutorialDivingIn)
from elixir import Entity, Field, OneToMany, ManyToOne, ManyToMany
from elixir import options_defaults, using_options, setup_all
# import some datatypes for table columns from Elixir
# (see http://www.sqlalchemy.org/docs/04/types.html for more)
from elixir import String, Unicode, Integer, DateTime, Float, Boolean
from turbogears.database import get_engine, metadata, session
from turbogears import identity

options_defaults.update(autosetup=False, metadata=metadata, session=session)


# your data model

class Doctor(Entity):

    using_options(order_by=['order', 'surname'])

    surname = Field(String(40))
    first_name = Field(String(40))
    last_name = Field(String(40))
    datetype = Field(Integer, default=1)
    visible = Field(Boolean, default=1)
    order = Field(Integer)
    norm = Field(Integer)

    def __repr__(self):
        return '<Doctor %s>' % (self.__str__(),)

    def __str__(self):
        return self.__unicode__().encode('utf-8')

    def __unicode__(self):
        return (u'%s %s.%s.' % (self.surname, self.first_name[:1],
            self.last_name[:1])).replace('  . .', '')

    submit = u"Сохранить"


class Organ(Entity):

    using_options(order_by=['order', 'name'])

    name = Field(String(40), unique=True)
    score = Field(Float)
    order = Field(Integer)
    visible = Field(Boolean, default=1)
    type = Field(Integer, default=0)

    def __repr__(self):
        return '<Organ %s (%i)>' % (self.__str__(), self.score)

    def __str__(self):
        return self.__unicode__().encode('utf-8')

    def __unicode__(self):
        return self.name

    submit = u"Сохранить"


class USG_Result(Entity):

    using_options(order_by=['date'])

    date = Field(DateTime, nullable=False, default=datetime.now)
    doctor = ManyToOne('Doctor')
    subresults = OneToMany('USG_SubResult')

    def __repr__(self):
        return '<USG_Result %s %r %r>' % (
            self.doctor.surname.encode('utf-8'),
            self.date, self.subresults)


class USG_SubResult(Entity):

    result = ManyToOne('USG_Result')
    organ = ManyToOne('Organ')
    number = Field(Integer)

    def __repr__(self):
        return '<USG_SubResult %s x %s>' % (
            self.organ.name.encode('utf-8'), self.number)


# the identity model

class Visit(Entity):
    """A visit to your site."""

    using_options(tablename='visit')

    visit_key = Field(String(40), primary_key=True)
    created = Field(DateTime, nullable=False, default=datetime.now)
    expiry = Field(DateTime)

    @classmethod
    def lookup_visit(cls, visit_key):
        return Visit.get(visit_key)
    by_visit_key = lookup_visit


class VisitIdentity(Entity):
    """A Visit that is linked to a User object."""

    using_options(tablename='visit_identity')

    visit_key = Field(String(40), primary_key=True)
    user = ManyToOne('User', colname='user_id', use_alter=True)

    @classmethod
    def by_visit_key(cls, visit_key):
        """Look up VisitIdentity by given visit key."""
        return cls.get(visit_key)


class Group(Entity):
    """An ultra-simple Group definition."""

    using_options(tablename='tg_group')

    group_id = Field(Integer, primary_key=True)
    group_name = Field(Unicode(16), unique=True, nullable=False)
    display_name = Field(Unicode(255))
    created = Field(DateTime, default=datetime.now)
    users = ManyToMany('User', tablename='user_group')
    permissions = ManyToMany('Permission', tablename='group_permission')

    def __repr__(self):
        return '<Group: name="%s", display_name="%s">' % (
            self.group_name, self.display_name)

    def __unicode__(self):
        return self.display_name or self.group_name

    @classmethod
    def by_group_name(cls, group_name):
        """Look up Group by given group name."""
        return cls.get_by(group_name=group_name)
    by_name = by_group_name


class User(Entity):
    """Reasonably basic User definition.

    Probably would want additional attributes.

    """

    using_options(tablename='tg_user')

    user_id = Field(Integer, primary_key=True)
    user_name = Field(Unicode(16), unique=True, nullable=False)
    email_address = Field(Unicode(255), unique=True)
    display_name = Field(Unicode(255))
    _password = Field(Unicode(40), colname='password')
    created = Field(DateTime, default=datetime.now)
    groups = ManyToMany('Group', tablename='user_group')

    def __repr__(self):
        return '<User: name="%s", email="%s", display name="%s">' % (
            self.user_name, self.email_address, self.display_name)

    def __unicode__(self):
        return self.display_name or self.user_name

    @property
    def permissions(self):
        p = set()
        for g in self.groups:
            p |= set(g.permissions)
        return p

    @classmethod
    def by_email_address(cls, email_address):
        """Look up User by given email address.

        This class method that can be used to search users based on their email
        addresses since it is unique.

        """
        return cls.get_by(email_address=email_address)

    @classmethod
    def by_user_name(cls, user_name):
        """Look up User by given user name.

        This class method that permits to search users based on their
        user_name attribute.

        """
        return cls.get_by(user_name=user_name)
    by_name = by_user_name

    def _set_password(self, password):
        """Run cleartext password through the hash algorithm before saving."""
        self._password = identity.encrypt_password(password)

    def _get_password(self):
        """Returns password."""
        return self._password

    password = property(_get_password, _set_password)


class Permission(Entity):
    """A relationship that determines what each Group can do."""

    using_options(tablename='permission')

    permission_id = Field(Integer, primary_key=True)
    permission_name = Field(Unicode(16), unique=True, nullable=False)
    description = Field(Unicode(255))
    groups = ManyToMany('Group', tablename='group_permission')

    def __repr__(self):
        return '<Permission: name="%s">' % self.permission_name

    def __unicode__(self):
        return self.permission_name

    @classmethod
    def by_permission_name(cls, permission_name):
        """Look up Permission by given permission name."""
        return cls.get_by(permission_name=permission_name)
    by_name = by_permission_name

# set up all Elixir entities declared above

setup_all()

# functions for populating the database

def bootstrap_model(clean=False, user=None):
    """Create all database tables and fill them with default data.

    This function is run by the 'bootstrap' function from the command module.
    By default it calls two functions to create all database tables for your
    model and optionally create a user.

    You can add more functions as you like to add more boostrap data to the
    database or enhance the functions below.

    If 'clean' is True, all tables defined by your model will be dropped before
    creating them again. If 'user' is not None, 'create_user' will be called
    with the given username.

    """
    create_tables(clean)
    if user:
        create_default_user(user)
    from uzi.prefill import prefill_tables
    prefill_tables()

def create_tables(drop_all=False):
    """Create all tables defined in the model in the database.

    Optionally drop existing tables before creating them.

    """
    get_engine()
    if drop_all:
        print "Dropping all database tables defined in model."
        from elixir import drop_all
        drop_all()
    from elixir import create_all
    create_all()

    print "All database tables defined in model created."

def create_default_user(user_name, password=None):
    """Create a default user."""
    try:
        u = User.by_user_name(user_name)
    except:
        u = None
    if u:
        print "User '%s' already exists in database." % user_name
        return
    from getpass import getpass
    from sys import stdin
    while password is None:
        try:
            password = getpass("Enter password for user '%s': "
                % user_name.encode(stdin.encoding)).strip()
            password2 = getpass("Confirm password: ").strip()
            if password != password2:
                print "Passwords do not match."
            else:
                password = password.decode(stdin.encoding)
                break
        except (EOFError, KeyboardInterrupt):
            print "User creation cancelled."
            return
    u = User()
    u.user_name = user_name
    u.display_name = u"Default User"
    u.email_address = u"%s@nowhere.xyz" % user_name
    u.password = password
    session.add(u)
    session.flush()
    print "User '%s' created." % user_name
